import math
import numpy as np


def circuit_dynamics(mdl, container_handle, caller_prop_handle=None):

    new_value = None
    prop_name = None
    if caller_prop_handle:
        new_value = mdl.get_property_disp_value(caller_prop_handle)
        prop_name = mdl.get_name(caller_prop_handle)

    # Properties routines:
    # ------------------------------------------------------------------------------------------------------------------
    #  "load_types" property code
    # ------------------------------------------------------------------------------------------------------------------
    if prop_name in ["load1_type", "load2_type", "load3_type",
                     "load4_type", "load5_type"]:

        comp_handle = mdl.get_parent(container_handle)
        dc_handle = mdl.get_item("DC", parent=comp_handle)
        load_name = prop_name.split("_")[0].title()
        load_num = int(load_name[-1])

        # Scada Input / Cte Vars
        enable_name = f"Enable {load_name}"
        enable_handle = mdl.get_item(enable_name, parent=comp_handle)
        enable_pos = None
        if enable_handle:
            enable_pos = mdl.get_position(enable_handle)
        # Bus Join Vars
        bus_join_en_handle = mdl.get_item("BusJoin_en", parent=comp_handle)
        if load_num == 1:
            bus_en_term_name = "in"
        else:
            bus_en_term_name = f"in{load_num-1}"
        if bus_join_en_handle:
            bus_en_term_handle = mdl.term(bus_join_en_handle, bus_en_term_name)
        # LUT Vars
        lut_name = f"LUT_{load_name}"
        lut_handle = mdl.get_item(lut_name, parent=dc_handle)
        lut_pos = None
        if lut_handle:
            lut_pos = mdl.get_position(lut_handle)
        bus_split_lut_handle = mdl.get_item("BusSplit_LUT", parent=dc_handle)
        if load_num == 1:
            bus_split_lut_term_name = "out"
        else:
            bus_split_lut_term_name = f"out{load_num - 1}"
        bus_split_term_handle = mdl.term(bus_split_lut_handle, bus_split_lut_term_name)
        lut_tag_handle = mdl.get_item(f"gotoNTC_{load_name}", parent=dc_handle, item_type="tag")

        # Actions
        if new_value == "Not Used":
            # Electrical Modifications
            disable_enable_load(mdl, container_handle, load_name.lower(), False)
            # Enable Logic
            mdl.delete_item(enable_handle)
            enable_handle = mdl.create_component("core/Constant",
                                                 name=enable_name,
                                                 parent=comp_handle,
                                                 position=enable_pos)
            mdl.set_property_value(mdl.prop(enable_handle, "execution_rate"), "ts")
            mdl.set_property_value(mdl.prop(enable_handle, "value"), "0")
            mdl.set_property_value(mdl.prop(enable_handle, "signal_type"), "uint")
            mdl.create_connection(bus_en_term_handle, mdl.term(enable_handle, "out"))
            # LUT logic
            mdl.delete_item(lut_handle)
            lut_handle = mdl.create_component("core/Gain",
                                              name=lut_name,
                                              parent=dc_handle,
                                              position=lut_pos,
                                              hide_name=True)
            mdl.set_property_value(mdl.prop(lut_handle, "gain"), "0")
            mdl.create_connection(bus_split_term_handle, mdl.term(lut_handle, "in"))
            mdl.create_connection(lut_tag_handle, mdl.term(lut_handle, "out"))
        else:
            # Electrical Modifications
            disable_enable_load(mdl, container_handle, load_name.lower(), True)
            # Enable logic
            mdl.delete_item(enable_handle)
            enable_handle = mdl.create_component("core/SCADA Input",
                                                 name=enable_name,
                                                 parent=comp_handle,
                                                 position=enable_pos)
            mdl.set_property_value(mdl.prop(enable_handle, "execution_rate"), "ts")
            mdl.set_property_value(mdl.prop(enable_handle, "max"), "1")
            mdl.set_property_value(mdl.prop(enable_handle, "min"), "0")
            mdl.set_property_value(mdl.prop(enable_handle, "signal_type"), "uint")
            mdl.create_connection(bus_en_term_handle, mdl.term(enable_handle, "out"))
            # LUT logic
            mdl.delete_item(lut_handle)
            lut_handle = mdl.create_component("core/1D look-up table",
                                              name=lut_name,
                                              parent=dc_handle,
                                              position=lut_pos,
                                              hide_name=True)
            mdl.set_property_value(mdl.prop(lut_handle, "in_vec_x"), f"xarray{load_num}")
            mdl.set_property_value(mdl.prop(lut_handle, "out_vec_f_x"), f"yarray{load_num}")
            mdl.create_connection(bus_split_term_handle, mdl.term(lut_handle, "addr"))
            mdl.create_connection(lut_tag_handle, mdl.term(lut_handle, "value"))


def mask_dialog_dynamics(mdl, container_handle, caller_prop_handle=None):
    new_value = None
    prop_name = None
    if caller_prop_handle:
        new_value = mdl.get_property_disp_value(caller_prop_handle)
        prop_name = mdl.get_name(caller_prop_handle)

    # Properties Routines
    # --------------------------------------------------------------------------
    # Load Type / Power
    # --------------------------------------------------------------------------
    if prop_name in ["load1_type", "load2_type", "load3_type",
                     "load4_type", "load5_type",
                     "load1_power", "load2_power", "load3_power",
                     "load4_power", "load5_power"]:

        load_name = prop_name.split("_")[0]
        prop_type = prop_name.split("_")[1]
        ct_names = ["r", "l", "cdc", "rdc"]
        ct_prop_names = ["_".join([load_name, cname]) for cname in ct_names]
        ct_prop_handles = [mdl.prop(container_handle, prop) for prop in ct_prop_names]
        power_prop_handle = mdl.prop(container_handle, load_name + "_power")

        if prop_type == "type":
            if new_value == "User Defined" :
                [mdl.enable_property(prop) for prop in ct_prop_handles]
                mdl.enable_property(power_prop_handle)
            elif new_value == "Not Used":
                [mdl.disable_property(prop) for prop in ct_prop_handles]
                mdl.disable_property(power_prop_handle)
            else:
                [mdl.disable_property(prop) for prop in ct_prop_handles]
                mdl.enable_property(power_prop_handle)

        coefs = get_load_constants(mdl, container_handle, load_name)
        # Not use scientific notation for resistances
        mdl.set_property_disp_value(ct_prop_handles[0], round(coefs[0], 5))
        mdl.set_property_disp_value(ct_prop_handles[3], round(coefs[3], 5))
        mdl.set_property_disp_value(ct_prop_handles[1], sc_notation(coefs[1]))
        mdl.set_property_disp_value(ct_prop_handles[2], sc_notation(coefs[2]))


def get_load_constants(mdl, container_handle, load_name):
    prop_type_handle = mdl.prop(container_handle, load_name + "_type")
    prop_power_handle = mdl.prop(container_handle, load_name + "_power")

    load_type = None
    load = mdl.get_property_disp_value(prop_type_handle)
    voltage = float(mdl.get_property_disp_value(mdl.prop(container_handle, "load_voltage")))
    try:
        power = float(mdl.get_property_disp_value(prop_power_handle)) * 1000
        power = 1e-3 if power == 0 else power
    except:
        power = 1e-3
    freq = float(mdl.get_property_disp_value(mdl.prop(container_handle, "load_freq")))
    # Classification
    if load == "Fluorescent Lamp":
        load_type = "CFL/LFL"
    elif load == "Microwave Oven":
        load_type = "SMPS_pPFC"
    elif load == "Low-Power Consumer Electronics":
        load_type = "SMPS_noPFC"
    elif load == "High-Power Consumer Electronics":
        load_type = "SMPS_pPFC"
    else:
        load_type = "Not Used"

    # Constants
    if load_type == "Not Used":
        R = 9999
        L = 9999
        Cdc = 9999
        Rdc = 100e3
    elif load_type == "SMPS_noPFC":
        Zbase = (voltage * voltage / power)
        R = (0.0017) * Zbase
        L = (6e-6) * Zbase / (2 * np.pi * freq)
        Cdc = 1 / (0.036 * Zbase * 2 * np.pi * freq)
        vdc = voltage * np.sqrt(2)
        Rdc = vdc * vdc / power
    elif load_type == "SMPS_pPFC":
        Zbase = (voltage * voltage / power)
        R = (0.0085) * Zbase
        L = (0.037) * Zbase / (2 * np.pi * freq)
        Cdc = 1 / (0.036 * Zbase * 2 * np.pi * freq)
        vdc = voltage * np.sqrt(2)
        Rdc = vdc * vdc / power
    elif load_type == "CFL/LFL":
        Zbase = (voltage * voltage / power)
        R = (2.16e-3) * Zbase
        L = (6e-6) * Zbase / (2 * np.pi * freq)
        Cdc = 1 / (0.25 * Zbase * 2 * np.pi * freq)
        vdc = voltage * np.sqrt(2)
        Rdc_ch = 0.01928 * vdc * vdc + 11.69 * vdc + 1110
        Rdc_disch = 24.5 * vdc + 94.29
        Rdc = sum([Rdc_ch, Rdc_disch]) / 2

    return [R, L, Cdc, Rdc]


def disable_enable_load(mdl, container_handle, load_name, action):
    comp_handle = mdl.get_parent(container_handle)
    dc_handle = mdl.get_item("DC", parent=comp_handle)
    load_num = load_name[-1]

    # DC
    dc_load_handle = mdl.get_item("PE - DCBus" + load_num, parent=dc_handle)
    if dc_load_handle:
        dc_load_pos = mdl.get_position(dc_load_handle)

    cmd_tag = mdl.get_item("cmd_tag" + load_num, parent=dc_handle, item_type="tag")
    cmd_tag_pos = mdl.get_position(cmd_tag)
    ntc_tag = mdl.get_item("ntc_tag" + load_num, parent=dc_handle, item_type="tag")
    ntc_tag_pos = mdl.get_position(ntc_tag)
    pos_tag = mdl.get_item("Load_pos" + load_num, parent=dc_handle, item_type="tag")
    neg_tag = mdl.get_item("Load_neg" + load_num, parent=dc_handle, item_type="tag")

    if action:

        # - Switch
        switch_term_handle = mdl.get_item("switch_term" + load_num, parent=dc_handle)
        if switch_term_handle:
            mdl.delete_item(switch_term_handle)

        dc_load_type = mdl.get_component_type_name(dc_load_handle)
        if dc_load_type == "el_open":
            mdl.delete_item(dc_load_handle)
            dc_load_handle = mdl.create_component("Residential Energy/PE Load - DCBus",
                                                  parent=dc_handle,
                                                  name="PE - DCBus" + load_num,
                                                  position=dc_load_pos)
            mdl.set_property_value(mdl.prop(dc_load_handle, "sw_r"), f"{load_name}_R/2")
            mdl.set_property_value(mdl.prop(dc_load_handle, "r_ac"), f"{load_name}_r/2")
            mdl.set_property_value(mdl.prop(dc_load_handle, "l_ac"), f"{load_name}_l")
            mdl.set_property_value(mdl.prop(dc_load_handle, "r_dc"), f"{load_name}_rdc")
            mdl.set_property_value(mdl.prop(dc_load_handle, "c_dc"), f"{load_name}_cdc")
            mdl.create_connection(cmd_tag, mdl.term(dc_load_handle, "sw"))
            mdl.create_connection(pos_tag, mdl.term(dc_load_handle, "DC+"))
            mdl.create_connection(neg_tag, mdl.term(dc_load_handle, "DC-"))

        # - NTC
        nct_term_handle = mdl.get_item("ntc_term" + load_num, parent=dc_handle)
        if nct_term_handle:
            mdl.delete_item(nct_term_handle)
            mdl.create_connection(ntc_tag, mdl.term(dc_load_handle, "ntc"))

    else:

        dc_load_type = mdl.get_component_type_name(dc_load_handle)
        if dc_load_type == "PE Load - DCBus":
            mdl.delete_item(dc_load_handle)
            dc_load_handle = mdl.create_component("core/Open Circuit",
                                                  parent=dc_handle,
                                                  name="PE - DCBus" + load_num,
                                                  position=dc_load_pos,
                                                  rotation="left")
            mdl.create_connection(pos_tag, mdl.term(dc_load_handle, "n_node"))
            mdl.create_connection(neg_tag, mdl.term(dc_load_handle, "p_node"))

        switch_term_handle = mdl.get_item("switch_term" + load_num, parent=dc_handle)
        if not switch_term_handle:
            switch_term_handle = mdl.create_component("core/Termination",
                                                      parent=dc_handle,
                                                      name="switch_term" + load_num,
                                                      position=(cmd_tag_pos[0]+64, cmd_tag_pos[1]))
            mdl.create_connection(cmd_tag, mdl.term(switch_term_handle, "in"))

        nct_term_handle = mdl.get_item("ntc_term" + load_num, parent=dc_handle)
        if not nct_term_handle:
            nct_term_handle = mdl.create_component("core/Termination",
                                                   parent=dc_handle,
                                                   name="ntc_term" + load_num,
                                                   position=(ntc_tag_pos[0] + 64, ntc_tag_pos[1]))
            mdl.create_connection(ntc_tag, mdl.term(nct_term_handle, "in"))


def sc_notation(val, num_decimals=2, exponent_pad=2):
    exponent_template = "{:0>%d}" % exponent_pad
    mantissa_template = "{:.%df}" % num_decimals

    order_of_magnitude = math.floor(math.log10(abs(val)))
    nearest_lower_third = 3 * (order_of_magnitude // 3)
    adjusted_mantissa = val * 10 ** (-nearest_lower_third)
    adjusted_mantissa_string = mantissa_template.format(adjusted_mantissa)
    adjusted_exponent_string = "+-"[nearest_lower_third < 0] + exponent_template.format(abs(nearest_lower_third))
    return adjusted_mantissa_string + "e" + f"{int(adjusted_exponent_string)}"


def get_lut_values(mdl, container_handle, load_name):
    pf = float(mdl.get_property_disp_value(mdl.prop(container_handle, load_name + "_r"))) / 2
    pi = pf * 40
    tau = 1

    x = np.arange(0, tau * 10, 0.5)
    x = x.tolist()
    y = [0] * len(x)
    for cnt, value in enumerate(x):
        y[cnt] = pf + (pi - pf) * math.exp(-value / tau)

    return x, y


def get_pos_constants(load_num):
    offset = int(load_num) - 1
    base_pos = {"rdc_pos": (8304 + 272 * offset, 8136),
                "cdc_pos": (8400 + 272 * offset, 8136),
                "vmeas_pos": (8496 + 272 * offset, 8136),
                "switch_pos": (8400 + 272 * offset, 8024),
                "ntc_pos": (8400 + 272 * offset, 7888)
                }

    return base_pos


def define_icon(mdl, container_handle):
    """
    Defines the component icon based on its type

    :param mdl: Schematic API
    :param container_handle: Component Handle
    :return: no return
    """
    mdl.set_component_icon_image(container_handle, "./residential_comp_images/res_load.svg", rotate="no_rotate")
    mdl.set_color(container_handle, "blue")
    mdl.disp_component_icon_text(container_handle, "PE Based", size=10, relpos_x=0.5, relpos_y=0.93, rotate="no_rotate")
