from typhoon.api.schematic_editor.const import ITEM_COMPONENT, ITEM_PORT, ITEM_JUNCTION, ITEM_CONNECTION, ERROR_GENERAL
from typhoon.api.schematic_editor.exception import SchApiException, SchApiDuplicateConnectionException
import numpy as np


def dialog_dynamics(mdl, mask_handle):
    # --------------------------------------------------------------------------------
    # Description:
    #   This function changes the enabled status of properties and calucates property
    #   values when properties change.
    #
    # Inputs:
    #   prop: property
    #   new_value: string
    #
    # Outputs:
    #   None
    #
    # Author:
    #   Andy Sellers, April 2023
    # --------------------------------------------------------------------------------
    precision = 6
    
    model_definition_prop = mdl.prop(mask_handle, 'model_definition')
    model_definition = mdl.get_property_disp_value(model_definition_prop)
    
    R_prop = mdl.prop(mask_handle, "R")
    L_prop = mdl.prop(mask_handle, "L")
    Vb_prop = mdl.prop(mask_handle, "Vb")
    f_prop = mdl.prop(mask_handle, "f")
    pf_prop = mdl.prop(mask_handle, "pf")
    Sb_prop = mdl.prop(mask_handle, "Sb")
    x_r_ratio_prop = mdl.prop(mask_handle, "x_r_ratio")
        
    mdl.enable_property(Vb_prop)
    mdl.enable_property(f_prop)
    
    Vb = mdl.get_property_disp_value(Vb_prop)
    f = mdl.get_property_disp_value(f_prop)
    
    if model_definition == "S-U":
        mdl.disable_property(R_prop)
        mdl.disable_property(L_prop)
        mdl.enable_property(pf_prop)
        mdl.enable_property(Sb_prop)
        mdl.disable_property(x_r_ratio_prop)
        
        try:
            Sb = float(mdl.get_property_disp_value(Sb_prop))
            pf = float(mdl.get_property_disp_value(pf_prop))
            f = float(f)
            Vb = float(Vb)
            
            if Sb != 0:
                Z = Vb**2/Sb
                R = pf*Z
                mdl.set_property_disp_value(R_prop, np.round(R,precision))
                
                X = Z*(1-pf**2)**0.5
                if R != 0:
                    mdl.set_property_disp_value(x_r_ratio_prop, np.round(X/R,precision))
                else:
                    mdl.set_property_disp_value(x_r_ratio_prop, '-')
                
            else:
                X = None
                mdl.set_property_disp_value(R_prop, '-')
                mdl.set_property_disp_value(x_r_ratio_prop, '-')
            
            if (f != 0) and (X is not None):
                L = X/(2*np.pi*f)
                L_mH = L*1e3
                mdl.set_property_disp_value(L_prop, np.round(L_mH,precision))
            else:
                mdl.set_property_disp_value(L_prop, '-')
            
        except ValueError:
            mdl.set_property_disp_value(R_prop, '-')
            mdl.set_property_disp_value(L_prop, '-')
            mdl.set_property_disp_value(x_r_ratio_prop, '-')
            

    if model_definition == "S-U and X/R":
        mdl.disable_property(R_prop)
        mdl.disable_property(L_prop)
        mdl.disable_property(pf_prop)
        mdl.enable_property(Sb_prop)
        mdl.enable_property(x_r_ratio_prop)
        
        try:
            Sb = float(mdl.get_property_disp_value(Sb_prop))
            x_r_ratio = float(mdl.get_property_disp_value(x_r_ratio_prop))
            f = float(f)
            Vb = float(Vb)
            
            if Sb != 0:
                Z = Vb**2/Sb
                
                pf = (1+x_r_ratio**2)**-0.5
                mdl.set_property_disp_value(pf_prop, np.round(pf,precision))
                
                R = pf * Z
                mdl.set_property_disp_value(R_prop, np.round(R,precision))
                
                X = x_r_ratio * R
                
                if f != 0:
                    L = X/(2*np.pi*f)
                    L_mH = L*1e3
                    mdl.set_property_disp_value(L_prop, np.round(L_mH,precision))
                else:
                    mdl.set_property_disp_value(L_prop, '-')

            else:
                mdl.set_property_disp_value(R_prop, '-')
                mdl.set_property_disp_value(L_prop, '-')
                mdl.set_property_disp_value(pf_prop, '-')

            
        except ValueError:
            mdl.set_property_disp_value(R_prop, '-')
            mdl.set_property_disp_value(L_prop, '-')
            mdl.set_property_disp_value(pf_prop, '-')
        

    if model_definition == "R-L":
        mdl.enable_property(R_prop)
        mdl.enable_property(L_prop)
        mdl.disable_property(pf_prop)
        mdl.disable_property(Sb_prop)
        mdl.disable_property(x_r_ratio_prop)
        
        try:
            R = float(mdl.get_property_disp_value(R_prop))
            L_mH = float(mdl.get_property_disp_value(L_prop))
            L = L_mH*1e-3
            f = float(f)
            Vb = float(Vb)
            
            X = 2*np.pi*f*L
            if R != 0:
                mdl.set_property_disp_value(x_r_ratio_prop, np.round(X/R,precision))
            else:
                mdl.set_property_disp_value(x_r_ratio_prop, '-')
            
            Z = (X**2 + R**2)**0.5
            if Z != 0:
                mdl.set_property_disp_value(pf_prop, np.round(R/Z,precision))
                mdl.set_property_disp_value(Sb_prop, np.round(Vb**2/Z,precision)) 
            else:
                mdl.set_property_disp_value(pf_prop, '-')
                mdl.set_property_disp_value(Sb_prop, '-')
            
        except ValueError:
            mdl.set_property_disp_value(x_r_ratio_prop, '-')
            mdl.set_property_disp_value(pf_prop, '-')
            mdl.set_property_disp_value(Sb_prop, '-')


def calculate_parameters(mdl, mask_handle):
    model_definition = mdl.get_property_value(mdl.prop(mask_handle, "model_definition"))
    Vb = mdl.get_property_value(mdl.prop(mask_handle, "Vb"))
    f = mdl.get_property_value(mdl.prop(mask_handle, "f"))
    Sb = mdl.get_property_value(mdl.prop(mask_handle, "Sb"))
    pf = mdl.get_property_value(mdl.prop(mask_handle, "pf"))
    x_r_ratio = mdl.get_property_value(mdl.prop(mask_handle, "x_r_ratio"))

    if model_definition == "S-U" or model_definition == "S-U and X/R":
        # Validation
        if Vb <= 0:
            mdl.error("Nominal line voltage must be positive.", kind="General error",
                      context=mdl.prop(mask_handle, "Vb"))
        if f <= 0:
            mdl.error("Nominal frequency must be positive.", kind="General error", context=mdl.prop(mask_handle, "f"))

        if Sb <= 0:
            mdl.error("Short circuit capacity must be positive.", kind="General error",
                      context=mdl.prop(mask_handle, "Sb"))

        if model_definition == "S-U":
            # Validation
            if pf < 0:
                mdl.error("Inductive power factor must be greater than or equal to 0.", kind="General error",
                          context=mdl.prop(mask_handle, "pf"))
            if pf > 1:
                mdl.error("Inductive power factor must be less than or equal to 1.", kind="General error",
                          context=mdl.prop(mask_handle, "pf"))

            # Calculation
            L = ((Vb ** 2.0) / Sb) * ((1 - pf ** 2.0) ** 0.5) / (2 * np.pi * f)
            R = pf * (Vb ** 2.0) / Sb
            return L, R

        else:
            # Validation
            if x_r_ratio <= 0:
                mdl.error("X/R ratio must be positive.", kind="General error",
                          context=mdl.prop(mask_handle, "x_r_ratio"))
            # Calculation
            R = ((Vb ** 2.0) / Sb) / ((1 + x_r_ratio ** 2.0) ** 0.5)
            L = R * x_r_ratio / (2 * np.pi * f)
            return L, R


def pre_compile(mdl, item_handle):
    comp_handle = mdl.get_sub_level_handle(item_handle)
    model_definition = mdl.get_property_value(mdl.prop(item_handle, "model_definition"))

    if model_definition == "S-U" or model_definition == "S-U and X/R":
        L, R = calculate_parameters(mdl, item_handle)

        L1 = mdl.get_item("inductance_phase_A", parent=comp_handle, item_type=ITEM_COMPONENT)
        L2 = mdl.get_item("inductance_phase_B", parent=comp_handle, item_type=ITEM_COMPONENT)
        R1 = mdl.get_item("resistance_phase_A", parent=comp_handle, item_type=ITEM_COMPONENT)
        R2 = mdl.get_item("resistance_phase_B", parent=comp_handle, item_type=ITEM_COMPONENT)

        mdl.set_property_value(mdl.prop(L1, "inductance"), L*1e-3)
        mdl.set_property_value(mdl.prop(L2, "inductance"), L*1e-3)
        mdl.set_property_value(mdl.prop(R1, "resistance"), R)
        mdl.set_property_value(mdl.prop(R2, "resistance"), R)
